@@ -53,7 +53,10 @@ struct pci_driver te12xp_driver;
 
 int debug = 0;
 static int j1mode = 0;
-static int alarmdebounce = 0;
+static int alarmdebounce = 2500; /* LOF/LFA def to 2.5s AT&T TR54016*/
+static int losalarmdebounce = 2500; /* LOS def to 2.5s AT&T TR54016*/
+static int aisalarmdebounce = 2500; /* AIS(blue) def to 2.5s AT&T TR54016*/
+static int yelalarmdebounce = 500; /* RAI(yellow) def to 0.5s AT&T devguide */
 static int loopback = 0;
 static int t1e1override = -1;
 static int unchannelized = 0;
@@ -84,13 +87,12 @@ struct t1 *ifaces[WC_MAX_IFACES];
 spinlock_t ifacelock = SPIN_LOCK_UNLOCKED;
 
 struct t1_desc {
-	char *name;
-	int flags;
+	const char *name;
 };
 
-static struct t1_desc te120p = { "Wildcard TE120P", 0 };
-static struct t1_desc te122 = { "Wildcard TE122", 0 };
-static struct t1_desc te121 = { "Wildcard TE121", 0 };
+static const struct t1_desc te120p = {"Wildcard TE120P"};
+static const struct t1_desc te122 = {"Wildcard TE122"};
+static const struct t1_desc te121 = {"Wildcard TE121"};
 
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 20)
 static kmem_cache_t *cmd_cache;
@@ -258,7 +260,7 @@ inline void cmd_decipher_vpmadt032(struct t1 *wc, unsigned char *readchunk)
 	}
 }
 
-static int config_vpmadt032(struct vpmadt032 *vpm)
+static int config_vpmadt032(struct vpmadt032 *vpm, struct t1 *wc)
 {
 	int res, channel;
 	GpakPortConfig_t portconfig = {0};
@@ -347,12 +349,14 @@ static int config_vpmadt032(struct vpmadt032 *vpm)
 		return -1;
 	}
 
-	for (channel = 0; channel < MAX_CHANNELS_PER_SPAN; ++channel) {
+	vpm->companding = (TYPE_T1 == wc->spantype) ?
+				ADT_COMP_ULAW : ADT_COMP_ALAW;
+	for (channel = 0; channel < ARRAY_SIZE(vpm->curecstate); ++channel) {
 		vpm->curecstate[channel].tap_length = 0;
 		vpm->curecstate[channel].nlp_type = vpm->options.vpmnlptype;
 		vpm->curecstate[channel].nlp_threshold = vpm->options.vpmnlpthresh;
 		vpm->curecstate[channel].nlp_max_suppress = vpm->options.vpmnlpmaxsupp;
-		memcpy(&vpm->desiredecstate[channel], &vpm->curecstate[channel], sizeof(vpm->curecstate[channel]));
+		vpm->curecstate[channel].companding = vpm->companding;
 
 		vpm->setchanconfig_from_state(vpm, channel, &chanconfig);
 		if ((res = gpakConfigureChannel(vpm->dspid, channel, tdmToTdm, &chanconfig, &cstatus))) {
@@ -490,7 +494,6 @@ static void cmd_dequeue_vpmadt032(struct t1 *wc, unsigned char *writechunk, int
 			writechunk[CMD_BYTE(4, 2, 1)] = 0;
 		}
 	} else if (test_and_clear_bit(VPM150M_SWRESET, &vpm->control)) {
-		debug_printk(1, "Booting  VPMADT032\n");
 		for (x = 0; x < 7; x++) {
 			if (0 == x)  {
 				writechunk[CMD_BYTE(x, 0, 1)] = (0x8 << 4);
@@ -527,11 +530,6 @@ static void cmd_dequeue_vpmadt032(struct t1 *wc, unsigned char *writechunk, int
 	}
 #endif
 
-	/* Now let's figure out if we need to check for DTMF */
-	/* polling */
-	if (test_bit(VPM150M_ACTIVE, &vpm->control) && !whichframe && !(atomic_read(&wc->txints) % 100))
-		schedule_work(&vpm->work);
-
 #if 0
 	/* This may be needed sometime in the future to troubleshoot ADT related issues. */
 	if (test_bit(VPM150M_ACTIVE, &vpm->control) && !whichframe && !(atomic_read(&wc->txints) % 10000))
@@ -729,7 +727,7 @@ static void t1_configure_t1(struct t1 *wc, int lineconfig, int txlevel)
 	else
 		mytxlevel = txlevel - 4;
 	fmr1 = 0x9e; /* FMR1: Mode 0, T1 mode, CRC on for ESF, 2.048 Mhz system data rate, no XAIS */
-	fmr2 = 0x22; /* FMR2: no payload loopback, auto send yellow alarm */
+	fmr2 = 0x20; /* FMR2: no payload loopback, don't auto yellow alarm */
 	if (loopback)
 		fmr2 |= 0x4;
 
@@ -1178,6 +1176,11 @@ static int echocan_create(struct dahdi_chan *chan, struct dahdi_echocanparams *e
 	if (!wc->vpmadt032) {
 		return -ENODEV;
 	}
+
+	*ec = wc->ec[chan->chanpos - 1];
+	(*ec)->ops = &vpm150m_ec_ops;
+	(*ec)->features = vpm150m_ec_features;
+
 	return vpmadt032_echocan_create(wc->vpmadt032, chan->chanpos - 1,
 		ecp, p);
 }
@@ -1188,7 +1191,17 @@ static void echocan_free(struct dahdi_chan *chan, struct dahdi_echocan_state *ec
 	if (!wc->vpmadt032)
 		return;
 
-	vpmadt032_echocan_free(wc->vpmadt032, chan, ec);
+	vpmadt032_echocan_free(wc->vpmadt032, chan->chanpos - 1, ec);
+}
+
+static void set_span_devicetype(struct t1 *wc)
+{
+	strncpy(wc->span.devicetype, wc->variety, sizeof(wc->span.devicetype) - 1);
+
+#if defined(VPM_SUPPORT)
+	if (wc->vpmadt032)
+		strncat(wc->span.devicetype, " (VPMADT032)", sizeof(wc->span.devicetype) - 1);
+#endif
 }
 
 static int t1_software_init(struct t1 *wc)
@@ -1216,12 +1229,7 @@ static int t1_software_init(struct t1 *wc)
 	sprintf(wc->span.name, "WCT1/%d", num);
 	snprintf(wc->span.desc, sizeof(wc->span.desc) - 1, "%s Card %d", wc->variety, num);
 	wc->span.manufacturer = "Digium";
-	strncpy(wc->span.devicetype, wc->variety, sizeof(wc->span.devicetype) - 1);
-
-#if defined(VPM_SUPPORT)
-	if (wc->vpmadt032)
-		strncat(wc->span.devicetype, " with VPMADT032", sizeof(wc->span.devicetype) - 1);
-#endif
+	set_span_devicetype(wc);
 
 	snprintf(wc->span.location, sizeof(wc->span.location) - 1,
 		"PCI Bus %02d Slot %02d", dev->bus->number, PCI_SLOT(dev->devfn) + 1);
@@ -1302,7 +1310,7 @@ static void setchanconfig_from_state(struct vpmadt032 *vpm, int channel, GpakCha
 
 	chanconfig->PcmInPortA = 3;
 	chanconfig->PcmInSlotA = (channel + 1) * 2;
-	chanconfig->PcmOutPortA = 2;
+	chanconfig->PcmOutPortA = SerialPortNull;
 	chanconfig->PcmOutSlotA = (channel + 1) * 2;
 	chanconfig->PcmInPortB = 2;
 	chanconfig->PcmInSlotB = (channel + 1) * 2;
@@ -1377,14 +1385,13 @@ static int t1_hardware_post_init(struct t1 *wc)
 		options.vpmnlptype = vpmnlptype;
 		options.vpmnlpthresh = vpmnlpthresh;
 		options.vpmnlpmaxsupp = vpmnlpmaxsupp;
+		options.channels = (wc->spantype == TYPE_T1) ? 24 : 32;
 
 		wc->vpmadt032 = vpmadt032_alloc(&options, wc->name);
 		if (!wc->vpmadt032)
 			return -ENOMEM;
 
-		wc->vpmadt032->context = wc;
 		wc->vpmadt032->setchanconfig_from_state = setchanconfig_from_state;
-		wc->vpmadt032->span = &wc->span;
 
 		res = vpmadt032_init(wc->vpmadt032, wc->vb);
 		if (res) {
@@ -1393,8 +1400,9 @@ static int t1_hardware_post_init(struct t1 *wc)
 			return -EIO;
 		}
 
-		config_vpmadt032(wc->vpmadt032);
+		config_vpmadt032(wc->vpmadt032, wc);
 
+		set_span_devicetype(wc);
 		module_printk("VPM present and operational (Firmware version %x)\n", wc->vpmadt032->version);
 		wc->ctlreg |= 0x10; /* turn on vpm (RX audio from vpm module) */
 		if (vpmtsisupport) {
@@ -1481,16 +1489,48 @@ static inline void t1_check_alarms(struct t1 *wc)
 			alarms |= DAHDI_ALARM_NOTOPEN;
 	}
 
-	if (c & 0xa0) {
-		if (wc->alarmcount >= alarmdebounce) {
-			if (!unchannelized)
-				alarms |= DAHDI_ALARM_RED;
-		} else
+	if (c & 0x20) { /* LOF/LFA */
+		if (wc->alarmcount >= (alarmdebounce/100))
+			alarms |= DAHDI_ALARM_RED;
+		else {
+			if (unlikely(debug && !wc->alarmcount)) {
+				/* starting to debounce LOF/LFA */
+				t1_info(wc, "LOF/LFA detected but "
+					"debouncing for %d ms\n",
+					alarmdebounce);
+			}
 			wc->alarmcount++;
+		}
 	} else
 		wc->alarmcount = 0;
-	if (c & 0x4)
-		alarms |= DAHDI_ALARM_BLUE;
+
+	if (c & 0x80) { /* LOS */
+		if (wc->losalarmcount >= (losalarmdebounce/100))
+			alarms |= DAHDI_ALARM_RED;
+		else {
+			if (unlikely(debug && !wc->losalarmcount)) {
+				/* starting to debounce LOS */
+				t1_info(wc, "LOS detected but debouncing "
+					"for %d ms\n", losalarmdebounce);
+			}
+			wc->losalarmcount++;
+		}
+	} else
+		wc->losalarmcount = 0;
+
+	if (c & 0x40) { /* AIS */
+		if (wc->aisalarmcount >= (aisalarmdebounce/100))
+			alarms |= DAHDI_ALARM_BLUE;
+		else {
+			if (unlikely(debug && !wc->aisalarmcount)) {
+				/* starting to debounce AIS */
+				t1_info(wc, "AIS detected but debouncing "
+					"for %d ms\n", aisalarmdebounce);
+			}
+			wc->aisalarmcount++;
+		}
+	} else
+		wc->aisalarmcount = 0;
 
 	/* Keep track of recovering */
 	if ((!alarms) && wc->span.alarms) 
@@ -1511,9 +1551,26 @@ static inline void t1_check_alarms(struct t1 *wc)
 		t1_setreg_full(wc, 0x20, fmr4 & ~0x20, NOT_VPM);
 		wc->flags.sendingyellow = 0;
 	}
-	
+	/*
 	if ((c & 0x10) && !unchannelized)
 		alarms |= DAHDI_ALARM_YELLOW;
+	*/
+
+	if ((c & 0x10) && !unchannelized) { /* receiving yellow (RAI) */
+		if (wc->yelalarmcount >= (yelalarmdebounce/100))
+			alarms |= DAHDI_ALARM_YELLOW;
+		else {
+			if (unlikely(debug && !wc->yelalarmcount)) {
+				/* starting to debounce AIS */
+				t1_info(wc, "yelllow (RAI) detected but "
+					"debouncing for %d ms\n",
+					yelalarmdebounce);
+			}
+			wc->yelalarmcount++;
+		}
+	} else
+		wc->yelalarmcount = 0;
+
 	if (wc->span.mainttimer || wc->span.maintstat) 
 		alarms |= DAHDI_ALARM_LOOPBACK;
 	wc->span.alarms = alarms;
@@ -1527,7 +1584,8 @@ static void handle_leds(struct t1 *wc)
 
 	led = wc->ledstate;
 
-	if (wc->span.alarms & (DAHDI_ALARM_RED | DAHDI_ALARM_BLUE)) {
+	if ((wc->span.alarms & (DAHDI_ALARM_RED | DAHDI_ALARM_BLUE))
+		|| wc->losalarmcount) {
 		/* When we're in red alarm, blink the led once a second. */
 		if (time_after(jiffies, wc->blinktimer)) {
 			led = (led & __LED_GREEN) ? SET_LED_RED(led) : UNSET_LED_REDGREEN(led);
@@ -1679,13 +1737,13 @@ static void timer_work_func(struct work_struct *work)
 {
 	struct t1 *wc = container_of(work, struct t1, timer_work);
 #endif
-	/* Called once every 100ms */
+	/* Called once every 100 ms */
 	if (unlikely(!test_bit(INITIALIZED, &wc->bit_flags)))
 		return;
 	t1_do_counters(wc);
 	t1_check_alarms(wc);
 	t1_check_sigbits(wc);
-	mod_timer(&wc->timer, jiffies + HZ/5);
+	mod_timer(&wc->timer, jiffies + HZ/10);
 }
 
 static void
@@ -1702,7 +1760,6 @@ static int __devinit te12xp_init_one(struct pci_dev *pdev, const struct pci_devi
 	struct t1_desc *d = (struct t1_desc *) ent->driver_data;
 	unsigned int x;
 	int res;
-	int startinglatency;
 	unsigned int index = -1;
 
 	for (x = 0; x < sizeof(ifaces) / sizeof(ifaces[0]); x++) {
@@ -1717,7 +1774,6 @@ static int __devinit te12xp_init_one(struct pci_dev *pdev, const struct pci_devi
 		return -EIO;
 	}
 	
-retry:
 	if (!(wc = kmalloc(sizeof(*wc), GFP_KERNEL))) {
 		return -ENOMEM;
 	}
@@ -1757,13 +1813,12 @@ retry:
 		return res;
 	}
 	
-	/* Keep track of which device we are */
-	pci_set_drvdata(pdev, wc);
 	if (VOICEBUS_DEFAULT_LATENCY != latency) {
 		voicebus_set_minlatency(wc->vb, latency);
 	}
+
+	voicebus_lock_latency(wc->vb);
 	voicebus_start(wc->vb);
-	startinglatency = voicebus_current_latency(wc->vb);
 	t1_hardware_post_init(wc);
 
 	for (x = 0; x < (wc->spantype == TYPE_E1 ? 31 : 24); x++) {
@@ -1783,31 +1838,14 @@ retry:
 
 	mod_timer(&wc->timer, jiffies + HZ/5);
 	t1_software_init(wc);
-	if (voicebus_current_latency(wc->vb) > startinglatency) {
-		/* The voicebus library increased the latency during
-		 * initialization because the host wasn't able to service the
-		 * interrupts from the adapter quickly enough.  In this case,
-		 * we'll increase our latency and restart the initialization.
-		 */
-		printk(KERN_NOTICE "%s: Restarting board initialization " \
-		 "after increasing latency.\n", wc->name);
-		latency = voicebus_current_latency(wc->vb);
-		dahdi_unregister(&wc->span);
-		voicebus_release(wc->vb);
-		wc->vb = NULL;
-		free_wc(wc);
-		wc = NULL;
-		goto retry;
-	}
-
 	module_printk("Found a %s\n", wc->variety);
-
+	voicebus_unlock_latency(wc->vb);
 	return 0;
 }
 
 static void __devexit te12xp_remove_one(struct pci_dev *pdev)
 {
-	struct t1 *wc = pci_get_drvdata(pdev);
+	struct t1 *wc = voicebus_pci_dev_to_context(pdev);
 #ifdef VPM_SUPPORT
 	unsigned long flags;
 	struct vpmadt032 *vpm = wc->vpmadt032;
@@ -1863,7 +1901,7 @@ static int __init te12xp_init(void)
 
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 23)
 	cmd_cache = kmem_cache_create(THIS_MODULE->name, sizeof(struct command), 0,
-#if LINUX_VERSION_CODE == KERNEL_VERSION(2, 6, 22)
+#if defined(CONFIG_SLUB) && (LINUX_VERSION_CODE == KERNEL_VERSION(2, 6, 22))
 				SLAB_HWCACHE_ALIGN | SLAB_STORE_USER, NULL, NULL);
 #else
 				SLAB_HWCACHE_ALIGN, NULL, NULL);
@@ -1896,6 +1934,11 @@ module_param(loopback, int, S_IRUGO | S_IWUSR);
 module_param(t1e1override, int, S_IRUGO | S_IWUSR);
 module_param(j1mode, int, S_IRUGO | S_IWUSR);
 module_param(alarmdebounce, int, S_IRUGO | S_IWUSR);
+#if 0 /* Don't export these parameters by default on the 2.2 branch */
+module_param(losalarmdebounce, int, S_IRUGO | S_IWUSR);
+module_param(aisalarmdebounce, int, S_IRUGO | S_IWUSR);
+module_param(yelalarmdebounce, int, S_IRUGO | S_IWUSR);
+#endif
 module_param(latency, int, S_IRUGO | S_IWUSR);
 #ifdef VPM_SUPPORT
 module_param(vpmsupport, int, S_IRUGO | S_IWUSR);
