@@ -31,6 +31,7 @@
 #endif
 
 #include <linux/kernel.h>
+#include <linux/sched.h>
 #include <linux/errno.h>
 #include <linux/module.h>
 #include <linux/device.h>
@@ -44,7 +45,7 @@
 #include "xpp_dahdi.h"
 #include "parport_debug.h"
 
-static const char rcsid[] = "$Id: xpp_dahdi.c 6642 2009-05-19 16:09:06Z tzafrir $";
+static const char rcsid[] = "$Id: xpp_dahdi.c 7493 2009-11-05 12:06:34Z tzafrir $";
 
 #ifdef CONFIG_PROC_FS
 struct proc_dir_entry *xpp_proc_toplevel = NULL;
@@ -940,6 +941,12 @@ static int xpp_hooksig(struct dahdi_chan *chan, enum dahdi_txsig txsig)
 			__FUNCTION__, pos);
 		return -ENODEV;
 	}
+	if(!xpd->xops->card_hooksig) {
+		LINE_ERR(xpd, pos,
+			"%s: No hooksig method for this channel. Ignore.\n",
+			__FUNCTION__);
+		return -ENODEV;
+	}
 	xbus = xpd->xbus;
 	BUG_ON(!xbus);
 	DBG(SIGNAL, "Setting %s to %s (%d)\n", chan->name, txsig2str(txsig), txsig);
@@ -1081,7 +1088,8 @@ int dahdi_register_xpd(xpd_t *xpd)
 	span->open = xpp_open;
 	span->close = xpp_close;
 	span->flags = DAHDI_FLAG_RBS;
-	span->hooksig = xpp_hooksig;	/* Only with RBS bits */
+	if(xops->card_hooksig)
+		span->hooksig = xpp_hooksig;	/* Only with RBS bits */
 	span->ioctl = xpp_ioctl;
 	span->maint = xpp_maint;
 	/*
